#   Copyright 2020 The KNIX Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import warnings
def deprecated(reason=""):
    def deprecator(func):
        def wrapper(*args,**kwargs):
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn("Call to deprecated function {}. {}".format(func.__name__, reason),
                          category=DeprecationWarning,
                          stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)  # reset filter
            return func(*args,**kwargs)
        return wrapper
    return deprecator





#   Copyright 2020 The KNIX Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import requests
import base64
import json
import random
import sys
import time
import logging

from .deprecated import deprecated


#logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class Execution(object):
    """ Execution represents the execution of a workflow that can be referenced by its execution ID
    an execution object is returned from asynchronous workflow invocations
    """
    def __init__(self, client, url, exec_id):
        self.client=client
        self.url=url
        self.execution_id = exec_id

    def get(self, timeout=60):
        try:
            r = self.client._s.post(self.url,
                params = {"executionId": self.execution_id},
                timeout=timeout)
        except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            raise Exception("Retrieving result of workflow from URL '"+self.url+"' failed due to "+type(e).__name__).with_traceback(sys.exc_info()[2])
        r.raise_for_status()
        return r.json()


class Workflow(object):
    """ Workflow represents a registered workflow, every method invocation or property assignment results in one or more calls to management functions
    """

    def __init__(self,client,wf):
        self.client=client
        self.id=wf["id"]
        self._name=wf["name"]
        self._modified=wf["modified"]
        self._status=wf.get("status",None)
        self._endpoints=wf.get("endpoints",None)

        self._deployment_error = ""
        self._json=None

    def __str__(self):
        if self._status == "deployed":
            return f"{self.id} ({self._name}, status: {self._status}, endpoints: {self._endpoints})"
        else:
            return f"{self.id} ({self._name}, status: {self._status})"

    @property
    def name(self):
        # TODO: workflow name could have been updated, decide if we should fetch workflow status
        return self._name

    @name.setter
    def name(self,name):
        res = self.client.action('modifyWorkflow',{'workflow':{'id':self.id,'name':name,'runtime':self._runtime}})
        self._name = name

    @property
    def modified(self):
        # TODO: workflow modification date could have been updated, decide if we should fetch workflow status
        return self._modified

    @property
    def status(self):
        data = self.client.action('getWorkflows',{'workflow':{'id':self.id}})
        self._status = data['workflow']['status']
        if self._status == "deployed":
            self._endpoints = data['workflow']['endpoints']
        else:
            self._endpoints = None
        if self._status == 'failed' and "deployment_error" in data['workflow']:
            self._deployment_error = data['workflow']['deployment_error']
        return self._status

    def get_deployment_error(self):
        return self._deployment_error

    @property
    def endpoint(self):
        if self.status == 'deployed':
            return random.choice(self._endpoints)
        else:
            return None

    @property
    def endpoints(self):
        if self.status == 'deployed':
            return self._endpoints
        else:
            return None

    @property
    def json(self):
        if not self._json:
            data = self.client.action('getWorkflowJSON',{'workflow':{'id':self.id}})
            self._json = base64.b64decode(data['workflow']['json']).decode().replace("\r","\n")
        return self._json


    @json.setter
    def json(self,json):
        if json != self.json:
            self._json = json
            self.client.action('uploadWorkflowJSON',{'workflow':{'id':self.id,'json':base64.b64encode(self._json.encode()).decode()}})


    def deploy(self, timeout=None):
        """ deploy a workflow and optionally wait in linearly increasing multiples of 1000ms
        :timeout: By default returns after calling deploy on the workflow without waiting for it to be actually deployed.
            If timeout is set to a numeric <= 0, it waits indefinitely in intervals of 1000ms, 2000ms, 3000ms, ...
            If timeout is set to a numeric > 0, it waits for the workflow to be deployed in increasing multiples of 100ms, but no longer than the timeout. When the timeout expires and the workflow is not deployed, the function raises an Exception
        """
        s = self.status
        if s == 'deployed':
            log.debug("deploy: wf %s already deployed",self.name)
            return
        elif s == 'deploying':
            log.debug("deploy: wf %s already being deployed",self.name)
        elif s == 'failed':
            log.debug("deploy: wf %s cannot be deployed", self.name)
            log.debug("deployment error: %s", self._deployment_error)
        else:
            self.client.action('deployWorkflow',{'workflow':{'id':self.id}})

        # if timeout is None, do not wait but return immediately even if it's not yet deployed
        if timeout is None:
            return

        sleep = 1
        if timeout > 0:
            # if timeout > 0, wait in increasing intervals but raise Exception if it's not deployed until the timeout expires
            t = time.time()
            end = t + timeout
            while t < end:
                s = self.status
                if s == 'deployed' or s == 'failed':
                    print()
                    return
                print("Waiting for deployment to come online; passed so far: " + str(round(t-end+timeout, 2)) + " seconds", end=" \r")
                sys.stdout.flush()
                t = time.time()
                if sleep < (end-t):
                    time.sleep(sleep)
                    sleep += 1
                else:
                    time.sleep(max(0,end-t))
            raise Exception("Deployment attempt timed out (%d)"%timeout)
        else:
            # if timeout <=0, wait in increasing intervals until deployed, even if this means forever
            while True:
                s = self.status
                if s == 'deployed' or s == 'failed':
                    return
                time.sleep(sleep)
                sleep += 1


    def undeploy(self, timeout=None):
        """ undeploy a workflow and optionally wait in linearly increasing multiples of 100ms
        :timeout: By default returns after calling undeploy on the workflow without waiting for it to be actually undeployed.
            If timeout is set to a numeric <= 0, it waits indefinitely in intervals of 100ms, 200ms, 300ms, ...
            If timeout is set to a numeric > 0, it waits for the workflow to be undeployed in increasing multiples of 100ms, but no longer than the timeout. When the timeout expires and the workflow is not undeployed, the function raises an Exception
        """
        if self.status == 'undeployed':
            log.debug("undeploy: wf %s not deployed",self.name)
            return
        if self.status == 'undeploying':
            log.debug("undeploy: wf %s is already being undeployed",self.name)
        else:
            self.client.action('undeployWorkflow',{'workflow':{'id':self.id}})

        # if timeout is None, do not wait but return immediately even if it's not yet deployed
        if timeout is None:
            return

        sleep = 1
        if timeout > 0:
            # if timout > 0, wait in increasing intervals but raise Exception if it's not undeployed until the timeout expires
            end = timeout + time.time()
            t = 0
            while t < end:
                if self.status == 'undeployed':
                    return
                t = time.time()
                if sleep < (end-t):
                    time.sleep(sleep)
                    sleep += 1
                else:
                    time.sleep(end-t)
            raise Exception("Deployment attempt timed out (%d)"%timeout)
        else:
            # if timeout <=0, wait in increasing intervals until undeployed, even if this means forever
            while True:
                if self.status == 'undeployed':
                    return
                time.sleep(sleep)
                sleep += 1

    def execute_async(self,data,timeout=30):
        """ execute a workflow asynchronously and returns an Execution object

        The function delivers an event to the frontend and returns an Execution object. Note that the timeout here applies to the delivery of the event, another timeout can be used when fetching the result with the Execution.get(timeout) method
        see Execution execute_async execute_async.get()

        :param data: the event dictionary passed to the workflow
        :type data: dict()
        :param timeout: time in seconds to wait for the event delivery to complete, otherwise throws ReadTimeout
        :type timeout: int
        :return: an Execution object to fetch the result
        :rtype: Execution
        :raises requests.exceptions.HTTPError: when the HTTP request to deliver the event fails
        :raises requests.exceptions.ConnectionError: when the platform can not be reached
        :raises requests.exceptions.ReadTimeout: when reading the HTTP result of delivering the event times out
        :raises ValueError: in case the response is not JSON
        """
        if self._status != "deployed":
            raise Exception("Workflow not deployed: " + self.name)

        # we are already deployed and have the endpoints stored in self._endpoints
        url = random.choice(self._endpoints)

        try:
            r = self.client._s.post(url,
                params={'async':'True'},
                json=data,
                allow_redirects=False,
                timeout=timeout)
        except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            raise Exception("Asynchronous execution of workflow at URL '"+url+"' failed due to "+type(e).__name__)

        r.raise_for_status()
        exec_id = r.text
        return Execution(self.client, url, exec_id)

    def execute(self,data,timeout=60, check_duration=False):
        """ execute a workflow synchronously

        The function sends an event to the frontend and waits for the result in the HTTP response

        :param data: the event dictionary passed to the workflow
        :type data: dict()
        :param timeout: time in seconds to wait for the workflow to complete, otherwise throws ReadTimeout
        :type timeout: int
        :return: the result of the workflow execution
        :rtype: dict()

        :raises requests.exceptions.HTTPError: when the HTTP request to execute the workflow fails (e.g. 500 ServerError)
        :raises requests.exceptions.ConnectionError: when the platform can not be reached
        :raises requests.exceptions.ReadTimeout: when reading the HTTP response with the workflow result times out
        :raises ValueError: in case the response is not JSON
        """
        #if self.status != 'deployed':
        #    self.deploy(-1)
        if self._status != "deployed":
            raise Exception("Workflow not deployed: " + self.name)

        # we are already deployed and have the endpoints stored in self._endpoints
        url = random.choice(self._endpoints)
        try:
            #postdata = {}
            #postdata["value"] = json.dumps(data)
            #postdata = json.dumps(postdata)
            if check_duration:
                t_start = time.time()
            r = self.client._s.post(url,
                params={},
                json=data,
                #headers={'Content-Type':'application/json'},
                #data=postdata,
                #headers={'Content-Type':'application/x-www-form-urlencoded'},
                #data=postdata,
                timeout=timeout)
            if check_duration:
                t_total = (time.time() - t_start) * 1000.0
        except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            raise Exception("Execution of workflow '"+self.name+"' at URL '"+url+"' failed due to "+type(e).__name__)
        r.raise_for_status()
        if check_duration:
            return r.json(), t_total
        return r.json()


    def logs(self, clear=False, ts_earliest=0.0, num_lines=500):
        """ fetch logs of this workflow
        :clear: default=False; if True, the function calls clear_logs() before returning
        :returns: a dict {'exceptions':<str>,'progress':<str>,'log':<str>}
        """
        #print("earliest: " + str(ts_earliest))
        data = self.client.action('retrieveAllWorkflowLogs',{'workflow':{'id':self.id, 'ts_earliest': ts_earliest, 'num_lines': num_lines}})
        res = {'exceptions':base64.b64decode(data['workflow']['exceptions']).decode(),
               'progress':base64.b64decode(data['workflow']['progress']).decode(),
               'log':base64.b64decode(data['workflow']['log']).decode(),
                'timestamp': data['workflow']['timestamp']}
        if clear:
            self.clear_logs()
        return res

    def clear_logs(self):
        """ clear logs of this workflow """
        try:
            self.client.action('clearAllWorkflowLogs',{'workflow':{'id':self.id}})
        except requests.exceptions.HTTPError as e:
            e.strerror += "while trying to clearAllWorkflowLogs for wf '"+self.name+"'/"+self.id
            raise e

    def delete_logs(self):
        """ alias for clear_logs() """
        self.clear_logs()

    def get_functions(self):
        fnames = []
        wfjson = json.loads(self.json)
        if 'States' in wfjson:
            for sname,state in list(wfjson['States'].items()):
                if 'Resource' in state:
                    fnames.append(state['Resource'])
        elif 'grains' in wfjson:
            for gdict in wfjson['grains']:
                fnames.append(gdict['name'])
        functions = []
        for f in self.client.functions:
            if f._name in fnames:
                functions.append(f)
                fnames.remove(f.name)
        if len(fnames) > 0:
            log.warn("Could not find all functions of workflow %s, missing %s"%(self.name,str(fnames)))
        return functions



#   Copyright 2020 The KNIX Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys
import base64
import datetime
from zipfile import ZipFile
import logging

from .deprecated import deprecated


#logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class Function(object):
    """ Function object represents a registered function, every method execution and property assignment triggers one or more calls to management functions
    """

    def __init__(self,client,function):
        self.client=client
        self.id=function["id"]
        self._runtime=function["runtime"]
        self._name=function["name"]
        self._modified=function["modified"]
        self._code=None
        self._metadata=None
        self._zip=None

    def __str__(self):
        return f"{self.id} ({self._name}, {self._runtime})"

    @property
    def name(self):
        # TODO: function name could be updated, but there's no management service call that would fetch the function's metadata only
        return self._name

    @name.setter
    def name(self,name):
        res = self.client.action('modifyFunction',{'function':{'id':self.id,'name':name,'runtime':self._runtime}})
        self._name = name

    @property
    def runtime(self):
        # TODO: function name could be updated, but there's no management service call that would fetch the function's metadata only
        return self._runtime

    @runtime.setter
    def runtime(self,runtime):
        res = self.client.action('modifyFunction',{'function':{'id':self.id,'name':self._name,'runtime':runtime}})
        self._runtime = runtime

    @property
    def modified(self):
        # TODO: function modification date could have been updated
        return self._modified

    @property
    def source(self):
        ret = dict()
        if self._code is None:
            res = self.client.action('getFunctionZipMetadata',{'function':{'id':self.id}})
            self._metadata = res['function']['metadata']
            chunks = res['function']['chunks']
            data = self.client.action('getFunctionCode',{'function':{'id':self.id}})
            self._code = base64.b64decode(data['function']['code']).decode()
            #print("Downloading and checking function zip: " + self._name)
            if chunks > 0: # download code
                self._zip = []
                for c in range(chunks):
                    data = self.client.action('getFunctionZip',{'function':{'id':self.id,'chunk':c}})
                    #print("Got chunk number: " + str(c+1) + "/" + str(chunks), end=' \r')
                    sys.stdout.flush()
                    self._zip.append(data['function']['code'])
                #print()
        ret['code'] = self._code
        if self._zip is not None:
            ret['zip'] = self._zip
            ret['metadata'] = self._metadata
        return ret

    @source.setter
    def source(self,g):
        if not self._code:
            self.source
        if not ('code' in g or 'zip' in g):
            raise Exception("Need dict of code and/or zip to update the function sources")
        if 'zip' in g:
            #if cmp(self._zip,g['zip']) == 0:
            if type(self._zip) == type(g['zip']) and ((self._zip > g['zip']) - (self._zip < g['zip'])) == 0:
                log.info("%s - %s - Not updating matching zip", self.client.user, self._name)
            else:
                newZip=True
                self._zip = g['zip']
                size=len(self._zip)
                print("Uploading function zip: " + self._name)
                for c in range(size):
                    data = self.client.action('uploadFunctionCode',{'function':{'id':self.id,'chunk':c,'code':self._zip[c],'format':'zip'}})
                    print(str(round((c+1)*100.0/size,2)) + " %", end=' \r')
                    sys.stdout.flush()
                print()
        if 'metadata' in g or 'zip' in g:
            if ('metadata' in g and self._metadata != g['metadata']) or ('zip' in g and len(self._zip) != len(g['zip'])):
                self._metadata = g['metadata']
                res = self.client.action('uploadFunctionZipMetadata',{'function':{'id':self.id,'chunks':len(self._zip),'metadata':self._metadata,'format':'zipMetadata'}})
            else:
                log.info("%s - %s - Not updating matching metadata", self.client.user, self._name)
        if 'code' in g:
            if self._code == g['code']:
                log.info("%s - %s - Not updating matching code", self.client.user, self._name)
            else:
                self._code = g['code']
                code = self._code
                if isinstance(code, str):
                    code = code.encode()
                data = self.client.action('uploadFunctionCode',{'function':{'id':self.id,'code':base64.b64encode(code).decode(),'format':"text"}})

    @property
    def code(self):
        return self.source['code']

    @code.setter
    def code(self,code):
        self.source = {'code':code}

    def upload(self,filename):
        src = self.source
        chunks=[]
        with open(filename, 'rb') as zf:
            content = base64.b64encode(zf.read()).decode()
            for pos in range(0,len(content),1024*1024):
                chunks.append(content[pos:pos+1024*1024])
        # create metadata information on the ZIP file contents
        metadata = "Last uploaded Zip file: <b>%s</b><br><br><table border='1'>" % (filename.split('/')[-1])
        with ZipFile(filename, 'r') as zf:
            codefile = None
            for fn in zf.namelist():
                if fn.endswith(self._name+".py") and (codefile is None or len(fn) < len(codefile)):
                    codefile = fn

                try:
                    fi = zf.getinfo(fn)
                except KeyError:
                    log.error('ERROR: Did not find %s in zip file', fn)
                else:
                    metadata += "<tr>" \
                        "<td style='padding: 5px 5px 5px 5px;'>%s</td>" \
                        "<td style='padding: 5px 5px 5px 5px;'>%.2f kB</td>" \
                        "<td style='padding: 5px 5px 5px 5px;'>%s</td>" \
                        "</tr>" % (fn, fi.file_size/1024.0, datetime.datetime(*fi.date_time))
            if codefile is not None:
                log.info("Updating code with file '%s' from zip archive", codefile)
                with zf.open(codefile, 'r') as code:
                    sc = code.read()
                    src['code'] = sc
        metadata += "</table>"
        # upload the zip file
        src['zip'] = chunks
        src['metadata'] = base64.b64encode(metadata.encode()).decode()
        self.source = src

    @property
    def requirements(self):
        data = self.client.action('getFunctionRequirements',{'function':{'id':self.id}})
        reqs = data['function']['requirements'].replace(" ", "+")
        return base64.b64decode(reqs.encode()).decode()

    @requirements.setter
    def requirements(self,requirements):
        self.client.action('uploadFunctionRequirements',{
            'function':{
                'id':self.id,
                'requirements':base64.b64encode(requirements.encode()).decode()
            }
        })



#   Copyright 2020 The KNIX Authors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import requests
import os
import json
import time
import urllib.request, urllib.parse, urllib.error
from .function import Function
from .workflow import Workflow
import logging
logging.basicConfig()
#logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)
from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)


import warnings
def deprecated(reason=""):
    def deprecator(func):
        def wrapper(*args,**kwargs):
            print("Use of deprecated func "+func.__name__+" - "+reason)
            return func(*args,**kwargs)
        return wrapper
    return deprecator

        #warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        #warnings.warn("Call to deprecated function {}. {}".format(func.__name__, reason),
        #              category=DeprecationWarning,
        #              stacklevel=2)
        #warnings.simplefilter('default', DeprecationWarning)  # reset filter


class MfnClient(object):
    """ MfnClient is a client to the MicroFunctions frontend
    The client communicates with management functions (e.g. authenticate, CRUD functions and workflows, deploy workflows, fetch logs).
    It provides workflow and function objects to modify them (e.g. upload ZIP, modify code, fetch logs).
    """


    @classmethod
    def _proxies(cls):
        proxies=dict()
        if 'HTTP_PROXY' in os.environ:
            log.info("Found environment variable HTTP_PROXY")
            proxies['http']=os.environ['HTTP_PROXY']
        if 'http_proxy' in os.environ:
            log.info("Found environment variable http_proxy")
            proxies['http']=os.environ['http_proxy']
        if 'HTTPS_PROXY' in os.environ:
            log.info("Found environment variable HTTPS_PROXY")
            proxies['https']=os.environ['HTTPS_PROXY']
        if 'https_proxy' in os.environ:
            log.info("Found environment variable https_proxy")
            proxies['https']=os.environ['https_proxy']

        if len(proxies) == 0:
            proxies = None
        return proxies


    @classmethod
    def load_env(cls, config=dict()):
        """ Creates a MfnClient config from environment variable

        :MFN_URL: MicroFunctions URL in the form http[s]://<hostname>/
        :MFN_USER: MicroFunctions username, typically an email address
        :MFN_PASSWORD: password authenticating the user
        :MFN_NAME: optional user name to register the user if it doesn't exist
        """
        if 'MFN_NODE' in os.environ and not 'MFN_URL' in os.environ:
            raise NotImplementedError("Use of MFN_NODE is deprecated, please change to MFN_URL")

        if 'MFN_URL' in os.environ:
            config['mfn_url'] = os.environ['MFN_URL']
            log.info("Found environment variable MFN_URL="+os.environ['MFN_URL'])
        if 'MFN_USER' in os.environ:
            config['mfn_user'] = os.environ['MFN_USER']
            log.info("Found environment variable MFN_USER="+os.environ['MFN_USER'])
        if 'MFN_NAME' in os.environ:
            config['mfn_name'] = os.environ['MFN_NAME']
            log.info("Found environment variable MFN_NAME="+os.environ['MFN_NAME'])
        if 'MFN_PASSWORD' in os.environ:
            config['mfn_password'] = os.environ['MFN_PASSWORD']
            log.info("Found environment variable MFN_PASSWORD")
        if 'proxies' not in config:
            config['proxies'] = cls._proxies()
        return config


    @classmethod
    def load_json(cls, config=dict(), filename=None):
        files = ['../settings.json','~/settings.json','./settings.json']
        if filename is not None:
            files.append(filename)
        for filename in files:
            """ get the json file """
            if os.path.isfile(filename):
                json_data = {}
                with open(filename) as json_file:
                    try:
                        json_data = json.load(json_file)
                        log.info("Using settings file", filename)
                        config.update(json_data)
                    except Exception:
                        log.error("Could not load json from settings file "+filename)
                        raise
        return config


    def __init__(self,mfn_url=None,mfn_user=None,mfn_password=None,mfn_name=None,proxies=None):
        """ creates a MfnClient and requires parameters to connect and authenticate with the management functions

        :url: URL (protocol+host) of the MicroFunctions server, e.g. https://microfunctions.org
        :user: username, typically the email address, of the user, e.g. test@test.com
        :password: password (plaintext) string authenticating the user
        :name: optional name of the user to have it created (registered) if it doesn't exist
        :proxies: a dict in the form {'http':'http://<proxyaddress>:<proxyport>','https':'http://<proxyhost>:<proxyport>'}

        tries to load the parameters from config files/environment variables if not given
        """
        config = MfnClient.load_json()
        config = MfnClient.load_env(config)
        url = mfn_url or config.get('mfn_url')
        user = mfn_user or config.get('mfn_user')
        password = mfn_password or config.get('mfn_password')
        name = mfn_name or config.get('mfn_name',None)
        proxies = proxies or config.get('proxies',None)
        if not url.startswith('http'):
            log.warning("WARNING: please use a format 'http://<hostname>:<port>' or 'https://...' for the url")
            if url.endswith(':443'):
                url = 'https://'+url
            else:
                url = 'http://'+url
            log.warning("... adjusted url to "+url)
            time.sleep(1)
        if url.endswith("/"):
            url = url[:-1]
        log.info(f"Connecting as {user} to {url}")
        self._s = requests.Session()
        if proxies:
            self._s.proxies.update(proxies)
        self._s.max_redirects = 10

        self.baseurl = str(url)
        self.mgmturl= self.baseurl.rstrip("/")+"/management"
        self.user=user
        self.token=None
        self.store=""
        self.name=name
        self.password=password
        self._functions=[]
        self._workflows=[]
        try:
            self.login()
        except requests.exceptions.HTTPError as e:
            log.error(str(e))
            raise Exception("Error reaching "+self.mgmturl+". Aborting")
        except:
            if self.name:
                log.info("Error logging in, trying to sign up for %s (%s)", self.name, self.user)
                userinfo = {}
                userinfo["email"] = self.user
                userinfo["password"] = self.password
                userinfo["name"] = self.name
                data_to_send = {}
                data_to_send["action"] = "signUp"
                data_to_send["data"] = {}
                data_to_send["data"]["user"] = userinfo
                #postdata = {}
                #postdata["value"] = json.dumps(data_to_send)

                #postdata = json.dumps(postdata)

                signUp = self._s.post(self.mgmturl,#verify=False,
                        params={},
                        json=data_to_send)
                        #headers={'Content-Type':'application/json'},
                        #data=postdata)
                try:
                    signUp.raise_for_status()
                    resp = signUp.json()
                    if resp['status'] != 'success':
                        raise Exception("Error signing up with "+self.mgmturl)
                    log.debug("Client signed up")
                    self.login()
                except requests.exceptions.HTTPError as e:
                    log.error(str(e))
                    raise Exception("Error signing up with "+self.mgmturl)
            else:
                raise Exception("Error logging in at "+self.mgmturl+". Maybe check credentials?")
        log.debug("Client %s obtained token %s",self.user,self.token)

    def disconnect(self):
        if self._s is not None:
            self._s.close()

    def version(self):
        data = self.action('version')
        return data.get('message','')

    def login(self):
        userinfo = {}
        userinfo["email"] = self.user
        userinfo["password"] = self.password
        data_to_send = {}
        data_to_send["action"] = "logIn"
        data_to_send["data"] = {}
        data_to_send["data"]["user"] = userinfo

        #postdata = {}
        #postdata["value"] = json.dumps(data_to_send)

        #postdata = json.dumps(postdata)

        r = self._s.post(self.mgmturl,
                    params={},
                    json=data_to_send)
                    #headers={'Content-Type':'application/json'},
                    #data=postdata)
        r.raise_for_status()
        resp = r.json()
        if resp['status'] == 'success':
            self.token = resp['data']['token']
            self.store = self.baseurl+resp['data']['storageEndpoint']
        else:
            raise Exception("Error logging in at "+self.mgmturl)


    def delete_user(self):
        data = {}
        data["user"] = {}
        data["user"]["token"] = self.token

        self.action("deleteAccount", data)

    def action(self,action,data=None):
        if data is None:
            data = dict()
        if 'user' not in data:
            data['user'] = {'token':self.token}
        log.debug("%s: %s -> %s", self.user, action, str(data)[:256]+(str(data)[256:] and '...'))
        data_to_send = {}
        data_to_send["action"] = action
        data_to_send["data"] = data
        #postdata = {}
        #postdata["value"] = json.dumps(data_to_send)
        #postdata = json.dumps(postdata)
        r = self._s.post(self.mgmturl,#verify=False,
                    params={},
                    json=data_to_send)
                    #headers={'Content-Type':'application/json'},
                    #data=postdata)
        r.raise_for_status()
        log.debug("%s: %s <- %s", self.user, action, r.text[:256]+(r.text[256:] and '...'))
        resp = r.json()
        if resp.get('status','') != 'success':
            if resp.get('has_error',False):
                raise Exception(f"MicroFunctions Error for action {action}: {resp['error_type']}")
            elif resp.get('data', None):
                raise Exception(f"MicroFunctions Error for action {action}: {resp['data']['message']}")
            else:
                raise Exception(f"MicroFunctions Error for action {action}: {r.text}")
        return resp['data']

    @property
    def grains(self):
        print("Use of deprecated property grains - Grains have been renamed to functions, use client.functions instead")
        return self.functions

    @property
    def functions(self):
        data = self.action('getFunctions')
        newset = []
        for f in data['functions']:
            found=None
            for old in self._functions:
                if f['id'] == old.id:
                    found=old
            if found is None:
                newset.append(Function(self,f))
            else:
                found._name = f['name']
                found._modified = f['modified']
                newset.append(found)
        #del self._functions
        self._functions = newset
        return self._functions


    @deprecated(reason="Grains have been renamed to functions, use find_function(..) instead")
    def findGrain(self,name):
        return self.find_function(name)

    @deprecated(reason="Grains have been renamed to functions, use find_function(..) instead")
    def find_grain(self,name):
        return self.find_function(name)

    def find_function(self,name):
        res = []
        for f in self.functions:
            if f._name == name:
                return f
            if f._name.startswith(name):
                res.append(f)
        # If nothing was found, try function IDs
        if len(res) == 0:
            for f in self._functions:
                if f.id == name:
                    return f
                if f.id.startswith(name):
                    res.append(f)
        if len(res) == 0:
            raise Exception("Could not find function matching "+name)
        else:
            if(len(res) > 1):
                log.warning("Found multiple functions for search word "+name+", returning first result")
            return res[0]

    @deprecated(reason="Grains have been renamed to functions, use add_function(..) instead")
    def addGrain(self,name,runtime='Python 3.6'):
        return self.add_function(name,runtime)

    @deprecated(reason="Grains have been renamed to functions, use add_function(..) instead")
    def add_grain(self,name,runtime='Python 3.6'):
        return self.add_function(name, runtime)

    def add_function(self,name,runtime='Python 3.6'):
        """ add a function

        returns an existing function if the name exists, registers a new function name if it doesn't exist
        :name: name of the function
        """
        for f in self.functions:
            if f._name == name:
                return f
        data = self.action('addFunction',{'function':{'name':name,'runtime':runtime}})
        gd = data['function']
        f = Function(self,gd)
        self._functions.append(f)
        return f


    @deprecated(reason="Grains have been renamed to functions, use delete_function(..) instead")
    def delGrain(self,g):
        return self.delete_function(g)

    @deprecated(reason="Grains have been renamed to functions, use delete_function(..) instead")
    def delete_grain(self,g):
        return self.delete_function(g)

    def delete_function(self,function):
        self._functions.remove(function)
        self.action('deleteFunction',{'function':{'id':function.id}})


    @property
    def workflows(self):
        data = self.action('getWorkflows')
        newset = []
        for wf in data['workflows']:
            found=None
            for old in self._workflows:
                if wf['id'] == old.id:
                    found=old
            if found is None:
                newset.append(Workflow(self,wf))
            else:
                newset.append(found)
        #del self._workflows
        self._workflows = newset
        return self._workflows


    @deprecated(reason="method name doesn't conform to PEP-8, use add_workflow(..) instead")
    def addWorkflow(self,name):
        return self.add_workflow(name)


    @deprecated(reason="method name doesn't conform to PEP-8, use delete_workflow(..) instead")
    def delWorkflow(self,wf):
        return self.delete_workflow(wf)

    def find_workflow(self,name):
        res = []
        for wf in self.workflows:
            if wf._name == name:
                return wf
            elif wf._name.startswith(name):
                res.append(wf)
        # If nothing was found, try workflow IDs
        if len(res) == 0:
            for wf in self.workflows:
                if wf.id == name:
                    return wf
                if wf.id.startswith(name):
                    res.append(wf)
        if len(res) == 0:
            raise Exception("Could not find workflow matching "+name)
        else:
            if(len(res) > 1):
                log.warning("Found multiple workflows for search word "+name+", returning first result")
            return res[0]


    def _get_state_names_and_resource(self, desired_state_type, wf_dict):
        state_list = []
        states = wf_dict['States']
        for state_name in states:
            state = states[state_name]
            state_type = state['Type']
            resource = ''
            if 'Resource' in state:
                resource = state['Resource']
            should_append = False
            if desired_state_type == 'all':
                should_append = True
            elif state_type == desired_state_type:
                should_append = True
            else:
                pass

            if should_append == True:
                state_list.append((state_type, state_name, resource))

            if state_type == 'Parallel':
                branches = state['Branches']
                for branch in branches:
                    parallel_state_list = self._get_state_names_and_resource(desired_state_type, branch)
                    state_list = state_list + parallel_state_list

            if state_type == 'Map':
                branch = state['Iterator']
                map_state_list = self._get_state_names_and_resource(desired_state_type, branch)
                state_list = state_list + map_state_list

        return state_list


    def add_workflow(self,name,filename=None):
        """ add a workflow

        returns an existing workflow if the name exists, registers a new workflow name if it doesn't exist
        :name: name of the workflow
        """
        for wf in self._workflows:
            if wf._name == name:
                return wf
        data = self.action('addWorkflow',{'workflow':{'name':name}})
        wfd = data['workflow']
        wf = Workflow(self,wfd)
        self._workflows.append(wf)

        if filename is not None:
            wfdesc = ""
            with open(filename,'r') as f:
                wfdesc = f.read()
            wfdir = os.path.dirname(os.path.abspath(filename))

            # set the WF json
            wf.json = wfdesc

            # parse the WF json to find required functions
            fnames = []
            wfjson = json.loads(wfdesc)
            if 'States' in wfjson:
                state_list = self._get_state_names_and_resource('Task', wfjson)
                for state_info in state_list:
                    #state_type = state_info[0]
                    #state_name = state_info[1]
                    state_resource = state_info[2]
                    fnames.append(state_resource)
            elif 'functions' in wfjson:
                for fdict in wfjson['functions']:
                    fnames.append(fdict['name'])

            for fname in fnames:
                fzipname = wfdir+"/%s.zip" % fname
                fpyname = wfdir+"/%s.py" % fname
                if not (os.path.exists(fzipname) or os.path.exists(fpyname)):
                    log.warn("Neither the ZIP file %s nor the source code %s was found for function %s" % (fzipname,fpyname,fname))

                log.info("Adding function: " + fname)

                f = self.add_function(fname)

                # Upload the .zip file
                if os.path.exists(fzipname):
                    f.upload(fzipname)

                # Upload the source code from the .py file
                if os.path.exists(fpyname):
                    with open(fpyname, 'r') as f:
                        fcode = f.read()
                    f.code = fcode

        return wf


    def delete_workflow(self,wf):
        if wf.status == 'deployed':
            wf.undeploy(0)
        if wf in self._workflows:
            self._workflows.remove(wf)
        self.action('deleteWorkflow',{'workflow':{'id':wf.id}})


    def keys(self, all_at_once=False):
        return self.list_keys(all_at_once=all_at_once)

    def keys(self):
        start=0
        step=100
        while start >= 0:
            data_to_send = {}
            data_to_send["action"] = "performStorageAction"
            data = {}
            user = {}
            user["token"] = self.token
            data["user"] = user
            storage = {}
            storage["action"] = "listKeys"
            storage["start"] = start
            storage["count"] = step
            data["storage"] = storage
            data_to_send["data"] = data
            r = self._s.post(self.mgmturl,
                    params={},
                    json=data_to_send)
            r.raise_for_status()
            if r.json()["status"] == "success":
                keylist = r.json()["data"]["keylist"]
                for key in keylist:
                    yield key
                if len(keylist) < step:
                    break
                start += step
            else:
                raise Exception("LISTKEYS failed: " + r.json()["data"]["message"])

    def list_keys(self, start=0, count=2000):
        data_to_send = {}
        data_to_send["action"] = "performStorageAction"
        data = {}
        user = {}
        user["token"] = self.token
        data["user"] = user
        storage = {}
        storage["action"] = "listKeys"
        storage["start"] = start
        storage["count"] = count
        data["storage"] = storage
        data_to_send["data"] = data
        r = self._s.post(self.mgmturl,
                params={},
                json=data_to_send)
        r.raise_for_status()
        if r.json()["status"] != "success":
            raise Exception("LISTKEYS failed: " + r.json()["data"]["message"])

        return r.json()["data"]["keylist"]


    def get(self, key):
        data_to_send = {}
        data_to_send["action"] = "performStorageAction"
        data = {}
        user = {}
        user["token"] = self.token
        data["user"] = user
        storage = {}
        storage["action"] = "getdata"
        storage["key"] = key
        data["storage"] = storage
        data_to_send["data"] = data
        r = self._s.post(self.mgmturl,
                params={},
                json=data_to_send)
        r.raise_for_status()
        if r.json()["status"] == "success":
            return r.json()["data"]["value"]
        else:
            raise Exception("GET failed: " + r.json()["data"]["message"])


    def put(self, key, value):
        data_to_send = {}
        data_to_send["action"] = "performStorageAction"
        data = {}
        user = {}
        user["token"] = self.token
        data["user"] = user
        storage = {}
        storage["action"] = "putdata"
        storage["key"] = key
        storage["value"] = value
        data["storage"] = storage
        data_to_send["data"] = data
        r = self._s.post(self.mgmturl,
                params={},
                json=data_to_send)
        r.raise_for_status()
        if r.json()["status"] != "success":
            raise Exception("PUT failed: " + r.json()["data"]["message"])


    def delete(self, key):
        data_to_send = {}
        data_to_send["action"] = "performStorageAction"
        data = {}
        user = {}
        user["token"] = self.token
        data["user"] = user
        storage = {}
        storage["action"] = "deletedata"
        storage["key"] = key
        data["storage"] = storage
        data_to_send["data"] = data
        r = self._s.post(self.mgmturl,
                params={},
                json=data_to_send)
        r.raise_for_status()
        if r.json()["status"] != "success":
            raise Exception("DELETE failed: " + r.json()["data"]["message"])
